/*
***************************************************************************************
 GPS Clock Using NEO6MV2 Module (Red Board with PPS)  Version 1.0   9/10/2021

 Copyright 2021
 Jed Margolin

 Uses gps-clock Rev A (Armageddon) board

 Uses MSP430G2553 because it has a UART. It has 16KB of program memory,512 bytes of RAM, and an ADC (not used here).
 An MSP430G2453 should also work. It has 8KB program memory, 512 bytes of RAM, and an ADC (not used here).
 An MSP430G2413 should also work. It has 8KB program memory, 512 bytes of RAM, No ADC (not used here anyway).

 Permission is granted to individuals to make copies for their personal use.
 Any commercial use requires my permission.

 Built with Code Composer Studio v10
 No crystal. Uses the factory DCO data

 Use with 0 - Register Optimization Only

 Use the switches with States to set UTC offset, DST, and 12/24 time
 state 0 -  Start the program with "# Sats in View"
 state 1 -  Wait until we get at least 5 satellites in View
 state 2 -  Get GPS and display local time; account for DST
 state 3 -  Display UTC
 state 4 -  Set Local Time 12/24 (Message)
 state 5 -  Set Local Time 12/24
 state 6 -  Leave Set Local Time 12/24
 state 7 -  Set UTC Offset
 state 8 -  Leave Set UTC Offset
 state 9 -  Set DST on/off (Message)
 state 10 - Set DST on/off
 state 11 - Display DST Start and End dates
 state 12 - Wait for Mode switch
 state 13 - Leave DST display
 state 14 - "# Sats in View"
 state 15 - Get and display # Satellites in View
 state 16 - Leave # Sats display
 state 17 - # Satellites in View Alarm
 state 18 - Leave # Satellites in View Alarm

 Port 1:
   D0  input     GPS PPS
   D1  input     UART RX
   D2  output    UART TX
   D3  input     INC SW
   D4  output    /Flag 1 (Seconds)
   D5  output    /Flag 2 (Minute End)
   D6  output    /Flag 3
   D7  output    /Flag 4 (Alarm)

 Port 2:
   D0  output    LCD D4
   D1  output    LCD D5
   D2  output    LCD D6
   D3  output    LCD D7
   D4  output    LCD /E (Enable)
   D5  output    LCD RS (Register Select)
   D6  input     MODE SW
   D7  input     DEC SW

***************************************************************************************
*/

#include <msp430.h>
#include "main.h"

int main(void)
{
   signed char data1, data2;

   WDTCTL = WDTPW + WDTHOLD;                 // Stop Watch Dog Timer

// No crystal. Use the factory default values for the DCO
   DCOCTL = CALDCO_16MHZ;      // 0x10F8 0x8A  factory calibration values
   BCSCTL1 = CALBC1_16MHZ;     // 0x10F9 0x8F

// Set up the Ports
// Port 1 - I/O
   P1SEL  &= ~(BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);
   P1SEL2 &= ~(BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);

// Port 1 - inputs
   P1REN |= (BIT0|BIT1|BIT3);    // pullup/pulldown enabled
   P1OUT |= (BIT0|BIT1|BIT3);    //
   P1DIR &= ~(BIT0|BIT1|BIT3);   // In

// Port 1 outputs
   P1DIR |= (BIT4|BIT5|BIT6|BIT7);     //Out
   P1OUT &= ~(BIT4|BIT5|BIT6|BIT7);

// Port 2 - I/O
   P2SEL  &= ~(BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);
   P2SEL2 &= ~(BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);

// Port 2 outputs
   P2DIR |= (BIT0|BIT1|BIT2|BIT3|BIT4|BIT5);     //Out
   P2OUT &= ~(BIT0|BIT1|BIT2|BIT3|BIT4|BIT5);

// Port 2 Inputs shared with xtal (which is not used here)
   P2DIR &= ~(BIT6|BIT7);   // In
   P2REN |= BIT6|BIT7;      // pullup/pulldown enabled
   P2OUT |= BIT6|BIT7;      //

// The MSP430G2553 has 3 ports. The PDIP brings out only 2 of them
// Port 3 - inputs
   P3SEL  &= ~(BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);  // In
   P3SEL2 &= ~(BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);
   P3REN |= (BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);    // pullup/pulldown enabled
   P3OUT |= (BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);    //
   P3DIR &= ~(BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);   // In

// Initialize the LCD
   lcd_reset();

   domsg(&msg1);        // "NEO6M GPS Clock"
   domsg(&msg2);        // " Jed Margolin"
   waitsec();
//---------------
// Do we need to initialize the Flash Memory Segment C with utc_offset, dst_observed and/or hours12?
// We are storing each parameter twice; unprogrammed data = -1 (0xFF)

   wr_flash_flag = 0;   // make sure we start out with the flag cleared

   data1 = INFOC[0];    // utc_offset
   data2 = INFOC[1];    // It is stored twice
   if( (data1 == -1) | (data2 == -1) | (data1 != data2) )  // unprogrammed or bad
   {
      utc_offset = -8;     // default: utc offset for Pacific time
      wr_flash_flag = 1;   // remember to write to it
   }
   else
   { utc_offset = data1; }

   data1 = INFOC[2];    // dst flag
   data2 = INFOC[3];    // It is stored twice
   if( (data1 == -1) | (data2 == -1) | (data1 != data2) )  // unprogrammed or bad
   {
      dst_observed = 1;   // default: dst is observed
      wr_flash_flag = 1;
   }
   else
   { dst_observed = data1; }

   data1 = INFOC[4];    // hours12 flag
   data2 = INFOC[5];    // It is stored twice
   if( (data1 == -1) | (data2 == -1) | (data1 != data2) )  // unprogrammed or bad
   {
      hours12 = 1;   // default: hours12 is enabled
      wr_flash_flag = 1;
   }
   else
   {
       hours12 = data1;
   }

   if(wr_flash_flag)    // If anything has changed write it into Flash Segment C
   {
      write_SegC(utc_offset, dst_observed, hours12);  // resets wr_flash_flag, increments count
   }
//---------------
   gps_msg_number = 1;    // Default is $GNZDA/$GPZDA (for time and date)

// Initialize UART for GPS module
   UCA0CTL1 |= UCSWRST;    // Hold in Reset

   P1SEL |= RXD + TXD;     // P1.1 and P2.2
   P1SEL2 |= RXD + TXD;

   UCA0CTL1 |= UCSSEL_2;   // SMCLK as UART Clock
   UCA0CTL0 = 0x00;        // no parity; LSB first; 8-bit data; one stop bit;UART mode; asynchronous
   UCA0BR0 = 0x83;         // Baud Rate LSB    16MHz/9600 = 1666.667 = 1666 = 0x0682
   UCA0BR1 = 0x06;         // Baud Rate MSB
   UCA0MCTL = UCBRS_6;     // Modulation UCBRSx = 6
   UCA0CTL1 &= ~UCSWRST;   // Initialize USCI state machine
   IE2 |= UCA0RXIE;        // Enable RX interrupt

// Set up timer Interrupts = 4ms  for reading switches
   CCTL0 = CCIE;                      // CCR0 interrupt enabled
   CCR0 = 63700;                      // 4ms interrupts (16,000,000 / 64,000 = 1,000  ->  63200
   TACTL = TASSEL_2 | ID_0 | MC_1;    // SMCLK, contmode = count up to CCR0
   TACTL |= TACLR;          // clear timer Resets TAR, clock divider, count direction

   __bis_SR_register(GIE);            // Interrupts without LPM0

   oldsec1 = 0;
   newsec1 = 0;
   oldsec2 = 0;
   newsec2 = 0;
   timeout = 0;
//================================================================
   while (1)           // main loop
   {
      while(!timer_flag)  { }   // wait for timer_flag to go high, 4ms interrupt
      timer_flag = 0;
//---------------
// Do the sounds, checking every 4ms so it doesn't get delayed. That is why this code is first.
// After Startup it might take up to a minute to synchronize with the time
   if(sound_enable) {do_sounds();}
//---------------
// Debounce and process the switches (on 4ms interrupts)
      do_switch(P2IN, BIT7, &switch1);         // PortIn, Bit mask, switch structure  INC
      do_switch(P1IN, BIT3, &switch2);         // PortIn, Bit mask, switch structure  DEC
      do_switch(P2IN, BIT6, &switch3);         // PortIn, Bit mask, switch structure  MODE
// We now know if a switch has been pressed

      if(switch3.press == 1)   // Mode switch
      {
          state++;
          if(state == 22)  {state = 3;}
          switch3.press = 0;   // reset the flag
      }

      (*state_ptr[state])();  // Do the main state machine. Some states first check to see if there is new GPS data (every second if there is any)
                              // Use the state number as the index to the list of pointers to the state functions
      asm(" nop");      // for emulator

      if(--timer_count == 0)    // 1 second (my flag, not GPS PPS)
      {
         timer_count = 250;
         my_second_flag = 1;
      }

   }  // end while main loop

}  // end main
//=======================================================
// Main State Machine

// Some cases first check to see if there is new GPS data (every second if there is any)
void state0()    // state 0 - start the program with "# Sats in View"
{
   blink = 0x01;         // Blink an arrow to show that it is working
   clear_lcd(0x80);      // First line of LCD
   clear_lcd(0xC0);      // Second line of LCD
   domsg(&msg19);        // "# Sats in View"

   gps_msg_number = 2;   // this data is in a different gps message (for getting #sats)

   state = 1;
}
//------------------
void state1()   // state 1 - Get and display # Satellites in View
{
   if(gotit)     // got the message
   {
      gotit = 0;

      if(blink)
      {
         lcd_wr_instr(0xC4);   // position
         lcd_wr_data(' ');     // blank
      }
      else
      {
         lcd_wr_instr(0xC4);   // position
         lcd_wr_data(0x7E);    // arrow
      }

      blink ^= 0x01;

      decode_data();           // decode the GPS data

      lcd_wr_instr(0xC6);      // position of # Sats in View data
      lcd_wr_data(satinv[0]);
      lcd_wr_data(satinv[1]);

      // stay in this mode here until there are at least 5 satellites in view
      // convert ASCII BCD to binary
      sats_in_view =  satinv[0] - 0x30;  // BCD most
      sats_in_view =  (sats_in_view<<3) + (sats_in_view<<1);  // first digit * 10 = *8 + *2
      sats_in_view = sats_in_view + (satinv[1] - 0x30);

      if(sats_in_view >= 5)
      {
          clear_lcd(0x80);      // First line of LCD
          clear_lcd(0xC0);      // Second line of LCD
          domsg(&msg26);        // "Waiting For PPS"

          state = 2;
      }
   }  // end of if(gotit)
}
//------------------
void state2()     // Wait For PPS
{
    newsec2 = P1IN & BIT0;                   // GPS PPS
    if( (newsec2 == 1) && (oldsec2 == 0) )   // got a transition
    {
       clear_lcd(0x80);     // First line of LCD
       gps_msg_number = 1;  // now we can get started
       sound_enable = 1;
       state = 3;
    }
    oldsec2 = newsec2;
}
//------------------
void state3()   // state 3 - Get GPS and display local time;
{
   if(gotit)     // got the new time and date (happens every second)
   {
       gotit = 0;
       IE2 &= ~UCA0RXIE;        // Disable RX interrupt

       decode_time();    // decode the time and date data

      // We have to add at least 1 second to the time because of the delay in writing to the display.
      // To do that we have to convert it to binary, so we might as well do that now.
      convert_utc_bin();

      if(utc_offset<0)
      {get_local_minus(utc_offset);}  // when UTC offset is negative
      else
      {get_local_plus(utc_offset);}   // when UTC offset is positive

      // Are we in DST? Assumes we have local date and time in binary
      if( dst_observed & is_dst() )
      {
          time_offset = utc_offset + 1;
          if(utc_offset<0)
          {get_local_minus(time_offset);}  // when UTC offset is negative
          else
          {get_local_plus(time_offset);}   // when UTC offset is positive
      }

      if(local_year != save_year)   // If the year has changed, get the new DST start and ends dates
      {
         find_dst(local_year);
         save_year = local_year;
      }

      IE2 |= UCA0RXIE;        // Enable RX interrupt

   }  // end of if(gotit)

   newsec2 = P1IN & BIT0;                   // GPS PPS
   if( (newsec2 == 1) && (oldsec2 == 0) )   // got a transition
   {
       display_local();  // Display local Time, adjust if 12 hour time
   }

   oldsec2 = newsec2;
}
//------------------
void state4()   // state 4 - Display UTC
{
   if(gotit)     // got the new time and date (happens every second)
   {
       gotit = 0;
       IE2 &= ~UCA0RXIE;        // Disable RX interrupt

       decode_time();   // decode the time and date data

// We have to add at least 1 second to the time because of the delay in writing to the display.
// To do that we have to convert it to binary
      convert_utc_bin();
      convert_utc_ascii();  // Convert the binary back to ascii
      domsg(&msg4);         // "UTC"
      IE2 |= UCA0RXIE;      // Enable RX interrupt
    }

   newsec2 = P1IN & BIT0;                   // GPS PPS
   if( (newsec2 == 1) && (oldsec2 == 0) )   // got a transition
   {
      display_utc();   // utc
   }

   oldsec2 = newsec2;

} // stay in this state until we get the message
//------------------
void state5()   // state 5 - Set Local Time 12/24 (Message)
{
   clear_lcd(0x80);   // First line of LCD
   clear_lcd(0xC0);   // Second line of LCD

   domsg(&msg6);  // "Local 12/24"};
   sound_enable = 0;
   state = 6;
} // display the message. Set the value in the next state
//------------------
void state6()  // state 6- Set Local Time 12/24
{
   if(hours12)  {domsg(&msg7);}  // 12
   else         {domsg(&msg8);}  // 24

   if(switch2.press == 1)
   {
      hours12 ^= 0x01;
      switch2.press = 0;   // reset the flag
      wr_flash_flag = 1;   // we changed something so write it to Flash (down the road)
   }

} // stay in this state until the MODE switch is pressed
//------------------
void state7()   // state 7- Leave Set Local Time 12/24
{
   clear_lcd(0x80);   // First line of LCD
   state = 8;
}
//------------------
void state8()   // state 8- Set UTC Offset
{
   unsigned char tempbcd[2];

   domsg(&msg5);  // "UTC Offset"
   if(utc_offset <0)
   {
      convert_hex_bcd2(-utc_offset,tempbcd);
      lcd_wr_instr(0x8C);   // position of utc offset
      lcd_wr_data('-');
      lcd_wr_data(tempbcd[0]);
      lcd_wr_data(tempbcd[1]);
   }
   else
   {
      convert_hex_bcd2(utc_offset, tempbcd);
      lcd_wr_instr(0x8C);   // position of utc offset
      lcd_wr_data('+');
      lcd_wr_data(tempbcd[0]);
      lcd_wr_data(tempbcd[1]);
   }

   if(switch2.press == 1)
   {
      if(utc_offset <=14)  {utc_offset++;}
      switch2.press = 0;   // reset the flag
      wr_flash_flag = 1;   // we changed something so write it to Flash soon
   }

   if(switch1.press == 1)
   {
      if(utc_offset > -12)  {utc_offset--;}
      switch1.press = 0;  // reset the flag
      wr_flash_flag = 1;    // we changed something so write it to Flash soon
   }

   time_offset = utc_offset + dst_hour;

} // stay in this state until the MODE switch is pressed
//------------------
void state9()   // state 9- Leave Set UTC Offset
{
   clear_lcd(0x80);   // First line of LCD
   state = 10;
}
//------------------
void state10()  // state 10- Set DST on/off (Message)
{
   clear_lcd(0x80);   // First line of LCD
   clear_lcd(0xC0);   // Second line of LCD

   domsg(&msg9);  // "DST"};
   state = 11;
}
//------------------
void state11()    // state 11 - Set DST on/off
{
   if(switch2.press == 1)
   {
      dst_observed ^= 0x01;
      switch2.press = 0;  // reset the flag
      wr_flash_flag = 1;  // we changed something so write it to Flash (down the road)
   }

   if(dst_observed)  {domsg(&msg10);}   // "Observed    "
   else              {domsg(&msg11);}   // "Not Observed"

}   // stay in this state until the MODE switch is pressed
//------------------
void state12()  // state 12 - Display DST Start and End dates
{
   unsigned char tempbcd[2];

   clear_lcd(0x80);   // First line of LCD
   clear_lcd(0xC0);   // Second line of LCD
   domsg(&msg17);     // "Start 3/"    (for dst start date)
   convert_hex_bcd2(dst_start_date,tempbcd);

   lcd_wr_instr(0x8C);  // position
   lcd_wr_data(tempbcd[0]);
   lcd_wr_data(tempbcd[1]);

   domsg(&msg18);       // "End  11/"    (for dst end date)
   convert_hex_bcd2(dst_end_date,tempbcd);

   lcd_wr_instr(0xCC);  // position
   lcd_wr_data(tempbcd[0]);
   lcd_wr_data(tempbcd[1]);

   state = 13;
}
//------------------
void state13()  // state 13 - wait for Mode switch
{
}
//------------------
void state14()  // state 14 - Leave DST display
{
   clear_lcd(0x80);   // First line of LCD
   clear_lcd(0xC0);   // Second line of LCD

   state = 15;

   // If we changed something write it to Flash
   if(wr_flash_flag)    // write it into Flash Segment C
   {
      write_SegC(utc_offset, dst_observed, hours12);  // resets wr_flash_flag, increments count
   }

}
//------------------
void state15()  // state 15 - "# Sats in View"
{
   oldsec2 = 0;
   newsec2 = 0;
   blink = 0x01;        // Blink an arrow to show that it is working
   clear_lcd(0x80);     // First line of LCD
   clear_lcd(0xC0);     // Second line of LCD
   domsg(&msg19);       // "# Sats in View"

   gps_msg_number = 2;  // this data is in a different gps message

   state = 16;
}
//------------------
void state16()  // state 16 - Get and display # Satellites in View
{
   newsec2 = my_second_flag;                // My flag for 1 second, not GPS PPS
   if( (newsec2 == 1) && (oldsec2 == 0) )  // get the transition
   {
      blink ^= 0x01;
      my_second_flag = 0;
   }
   oldsec2 = newsec2;

   if(gotit)     // got the new #sats
   {
      gotit = 0;

      if(blink)
      {
         lcd_wr_instr(0xC4);   // position
         lcd_wr_data(' ');     // blank
      }
      else
      {
         lcd_wr_instr(0xC4);   // position
         lcd_wr_data(0x7E);    // arrow
      }

      decode_data();           // decode the GPS data

      lcd_wr_instr(0xC6);      // position of # Sats in View data
      lcd_wr_data(satinv[0]);
      lcd_wr_data(satinv[1]);
   }

}
//------------------
void state17()  // state 17 - Leave # Sats display
{
    clear_lcd(0x80);   // First line of LCD
    clear_lcd(0xC0);   // Second line of LCD
    domsg(&msg20);     // "GPS Alarm Mode"

    state = 18;
}
//------------------
void state18()  // state 18 - # Satellites in View Alarm
{
   newsec2 = my_second_flag;                // My flag for 1 second, not GPS PPS
   if( (newsec2 == 1) && (oldsec2 == 0) )  // get the transition
   {
      blink ^= 0x01;
      my_second_flag = 0;
   }
   oldsec2 = newsec2;

   if(gotit)     // got the new time and date
   {
      timeout = 0;

      if(blink)
      {
         lcd_wr_instr(0xC4);   // position
         lcd_wr_data(' ');     // blank
      }
      else
      {
         lcd_wr_instr(0xC4);   // position
         lcd_wr_data(0x7E);    // arrow
      }

      decode_data();           // decode the GPS data

      lcd_wr_instr(0xC6);      // position of # Sats in View data
      lcd_wr_data(satinv[0]);
      lcd_wr_data(satinv[1]);

      sats_in_view =  satinv[0] - 0x30;  // BCD most
      sats_in_view =  (sats_in_view<<3) + (sats_in_view<<1);  // first digit * 10 = *8 + *2
      sats_in_view = sats_in_view + (satinv[1] - 0x30);

      if(sats_in_view >= 5)
      {
         domsg(&msg21);   // "Good"
         P1OUT &= ~BIT7;   // Flag 4
      }
      else
      {
         if(sats_in_view >= 4)
         {
            domsg(&msg22);   // "Ok"
            P1OUT &= ~BIT7;  // Flag 4
         }
         else
         {
            domsg(&msg23);  // "Oh Oh"}
            P1OUT |= BIT7;   // Flag 4
         }
      }
   }  // end of have gotten an update
   else   // have not gotten an update
   {
      timeout++;
      if(timeout >= 30000)   // 2 minutes without a signal
      {
          domsg(&msg23);     // "Oh Oh"}
          P1OUT |= BIT7;     // Flag 4
      }

   }

    gotit = 0;
}
//------------------
void state19()  // state 19 - Software Version
{
    clear_lcd(0x80);   // First line of LCD
    clear_lcd(0xC0);   // Second line of LCD

    domsg(&msg24);     // Software
    domsg(&msg25);     // Version x.y
    state = 20;
}
//------------------
void state20()  // state 20 - Wait for MODE switch
{
}
//------------------
void state21()  // state 21 - Back to the beginning
{
    clear_lcd(0x80);   // First line of LCD
    clear_lcd(0xC0);   // Second line of LCD

    gps_msg_number = 1;  // back to time/date gps message
    state = 3;
    sound_enable = 1;
}
//------------------
//==========================================================
// Timer A0 interrupt service routine
#if defined(__TI_COMPILER_VERSION__) || defined(__IAR_SYSTEMS_ICC__)
#pragma vector=TIMER0_A0_VECTOR
__interrupt void Timer_A (void)
#elif defined(__GNUC__)
void __attribute__ ((interrupt(TIMER0_A0_VECTOR))) Timer_A (void)
#else
#error Compiler not supported!
#endif
{
   timer_flag = 1;
}
//==========================================================
// UART RX Interrupt
#pragma vector = USCIAB0RX_VECTOR
__interrupt void USCIAB0RX_ISR(void)
{
   if(gps_msg_number == 1)   {get_gps_msg1();}  // $GNZDA/GPZDA  Time
   if(gps_msg_number == 2)   {get_gps_msg2();}  // $GPGSV  # of sats

   // ready for next data
   IFG2 &= ~UCA0RXIFG;      // clear the RX interrupt flag
}
//----------------------------
// For Time (GNZDA or GPZDA)
// For Time
void get_gps_msg1()
{
   static unsigned char state=0;
   unsigned char rxdata;

   rxdata = UCA0RXBUF;

   switch(state)
   {

      case 7 :   // Get the data for GPZDA/GNZDA

         gps_buffer[data_count] = rxdata;  // the data
         data_count++;

         if(rxdata == 0x0A)   // message ends with a LF
         {
            state = 0;
            data_count = 0;
            gotit = 1;       // got the packet
         }

         if(data_count == 32)
         {
            comm_error = 1;
            state = 0;
         }

         break;


      case 6 :

          if(rxdata == ',')
          {
             data_count = 0;
             state = 7;
          }
          else   // no comma, must be an error
          {
            state = 0;   // keep looking for '$'
          }

          break;

      case 5 :

          if(rxdata == 'A')
          {
             data_count = 0;
             state = 6;
          }
          else
          {
            state = 0;   // keep looking for '$'
          }

          break;

      case 4 :

          if(rxdata == 'D')
          {
             state = 5;
          }
          else
          {
            state = 0;   // keep looking for '$'
          }

          break;


      case 3 :

          if(rxdata == 'Z')
          {
             state = 4;
          }
          else
          {
            state = 0;   // keep looking for '$'
          }

          break;


      case 2 :

          if(  (rxdata == 'P') | (rxdata == 'N') )
          {
             state = 3;
          }
          else
          {
            state = 0;   // keep looking for '$'
          }

          break;

      case 1 :    // look for 'GNZDA,'  or 'GPZDA,'  (6 characters)

          if(rxdata == 'G')
          {
             state = 2;
          }
          else
          {
            state = 0;   // keep looking for '$'
          }

          break;

      case 0 :

         if(rxdata == '$')    // message starts with '$'
         {
            state = 1;
         }

         break;

   }  // end of switch
}  // message done
//--------------------------------------
// For # of sats
void get_gps_msg2()
{
   static unsigned char state=0;
   unsigned char rxdata;

   rxdata = UCA0RXBUF;

   switch(state)
   {

      case 2 :   // Get the data for GPGSV; we are only getting 8 bytes

        gps_buffer[data_count] = rxdata;  // the data
        data_count++;

        if(data_count == 8)
        {
           state = 0;
           gotit = 1;
        }

        break;

     case 1 :   // look for 'GPGSV,'  (6 characters)

        if(rxdata == GPSHEADER2[data_count])
        {
           data_count++;
           if(data_count == 6)    // We have the "GPGSV," header
           {
              state = 2;          // ready to get the data
              data_count = 0;
           }
        }
        else   // it isn't GNZDA
        {
          state = 0;   // keep looking
        }

        break;

     case 0 :

        if(rxdata == '$')    // message starts with '$'
        {
           state = 1;
           data_count = 0;
        }

        break;

   }  // end of switch

}
//==========================================================
// Decode GPS time from ZDA
void decode_time(void)
{
   unsigned char n;

   utc_hour[0] = gps_buffer[0];
   utc_hour[1] = gps_buffer[1];

   utc_min[0] = gps_buffer[2];
   utc_min[1] = gps_buffer[3];

   utc_sec[0] = gps_buffer[4];
   utc_sec[1] = gps_buffer[5];

   for(n=6; n<12; n++)
   {
      if(gps_buffer[n] == ',')
      {
         n++;
         break;
      }
   }
   utc_day[0] = gps_buffer[n++];
   utc_day[1] = gps_buffer[n++];
   n++;

   utc_month[0] = gps_buffer[n++];
   utc_month[1] = gps_buffer[n++];
   n++;

   utc_year[0] = gps_buffer[n++];
   utc_year[1] = gps_buffer[n++];
   utc_year[2] = gps_buffer[n++];
   utc_year[3] = gps_buffer[n];

   last_utc_sec[0] = utc_sec[0];
   last_utc_sec[1] = utc_sec[1];

}  // end of decode time and date
//===============================================
// Decode GPS data
// Example: $GPGSV,3,1,09,08,25,313,,10,71,326,28,15,15,045,,18,40,111,37*7F
void decode_data(void)
{
   unsigned int m,n;

//   IE2 &= !UCA0RXIE;        // disable RX interrupt for test

// the data we want is after the second ','
   for(n=0; n<8; n++)   // first ','
   {
      if(gps_buffer[n] == ',')
      {
         m = n+1;
         break;
      }
   }

   for(n=m; n<8; n++)   // second ','
   {
      if(gps_buffer[n] == ',')
      {
         break;
      }
   }

   satinv[0] = gps_buffer[n+1];
   satinv[1] = gps_buffer[n+2];

   asm(" nop");    // for emulator
}  // end of decode data
//==========================================================
void display_utc()   // utc time
{
   unsigned int weekday, m;
   unsigned char temp;

   lcd_wr_instr(0x80);        // position of time
   lcd_wr_data(utc_hour[0]);
   lcd_wr_data(utc_hour[1]);

   lcd_wr_data(':');
   lcd_wr_data(utc_min[0]);
   lcd_wr_data(utc_min[1]);

   lcd_wr_data(':');
   lcd_wr_data(utc_sec[0]);
   lcd_wr_data(utc_sec[1]);

   lcd_wr_instr(0xC0);        // position of date
   lcd_wr_data(utc_month[0]);
   lcd_wr_data(utc_month[1]);

   lcd_wr_data('/');
   lcd_wr_data(utc_day[0]);
   lcd_wr_data(utc_day[1]);

   lcd_wr_data('/');
   lcd_wr_data(utc_year[0]);
   lcd_wr_data(utc_year[1]);
   lcd_wr_data(utc_year[2]);
   lcd_wr_data(utc_year[3]);

   weekday = get_day_utc();     // day of the week
   lcd_wr_instr(0xCC);
   for(m=0;m<3;m++)
   {
      temp =  day_array[weekday][m];
      lcd_wr_data(temp);
   }

   asm(" nop");    // for emulator
}
//==================================================
//==================================================
// Get local time for negative UTC offset
// We already have year, month, day, hour in binary
void get_local_minus(signed char offset)   // minutes and seconds are the same
{
   unsigned char leap;

// This way we don't have to remember which ones stay the same
   local_hour = ut_hour;
   local_minute = ut_minute;
   local_second = ut_second;
   local_day    = ut_day;
   local_month  = ut_month;
   local_year   = ut_year;

   local_hour = local_hour + offset;    // utc hour + offset
   if(local_hour>= 0)  // Same day; we don't have to adjust days and maybe months and maybe years
   {
      asm(" nop");    // for emulator
      return;
    }
    else
    // now we have to decrement the day
    {
       // first add 24 to the hour since it is the previous utc day
       local_hour = local_hour + 24;

       // get the new day number
       local_day = local_day - 1;   // subtract a day

       if(local_day >= 1)    // we don't have to adjust months and maybe years
       {
          asm(" nop");    // for emulator
          return;
      }
      else  // shazbot, the day was zero so we have to adjust the month
      {
         // get the days in the previous month
         local_month = local_month -1;

         if(local_month == 0)  // roll back to the previous year
         {
            local_month = 12;
            local_year = local_year -1;  // and decrement the year
            leap = is_leap();            // is it a leap year?
            find_dst(local_year);
         }

         if(leap)
         {
            local_day = months_ly[local_month];  // the last day of the last month for regular years
         }
         else
         {
            local_day = months[local_month];    // the last day of the last month for regular years
         }

      }
      asm(" nop");    // for emulator
      return;
    }
// end of adjusting for local time which might have required adjusting
//    the day, month, and year
}
//=====================================================
// Get local time for positive UTC offset
// We already have year, month, day, hour in binary
void get_local_plus(signed char offset)   // minutes and seconds are the same
{
   unsigned char leap;

// This way we don't have to remember which ones stay the same
   local_hour = ut_hour;
   local_minute = ut_minute;
   local_second = ut_second;
   local_day    = ut_day;
   local_month  = ut_month;
   local_year   = ut_year;

   local_hour = local_hour + offset;    // utc hour + offset

   if(local_hour < 24)  // Same day; we don't have to adjust days and maybe months and maybe years
   {
      asm(" nop");    // for emulator
      return;
   }
   else
   // now we have to increment the day
   {
      // first subtract 24 to the hour since it is the next day
      local_hour = local_hour - 24;
      local_day = local_day + 1;     // add a day

      // First, is it a leap year?
      leap = is_leap();
      if(leap)
      {
         LastDay = months_ly[local_month];  // the last day of the month for leap years
      }
      else
      {
         LastDay = months[local_month];     // the last day of the month for regular years
      }

      if(day < LastDay)  // we don't have to adjust months and maybe years
      {
         asm(" nop");    // for emulator
         return;
      }
      else  // shazbot, the day was past the month so we have to adjust the month
      {
         local_day = 1;
         local_month = local_month + 1;

         if(local_month == 13)
         {
            local_month = 1;     // New year

            // and increment the year
            local_year = local_year + 1;
            leap = is_leap();
            find_dst(local_year);
         }
         else   // did not go to the next year so copy the utc year
         {
            return;
         }
         asm(" nop");    // for emulator
         return;
      }

    }
// end of adjusting for local time which might have required adjusting
//    the day, month, and year
}
//=====================================================
// Adjust if 12 hour time
void time12()
{
   if(hours12)
   {
      if(local_hour<12)
      {
         pmflag = 0;
         return;
      }

      if(local_hour == 0)  //12 AM
      {
         local_hour = 12;
         pmflag = 0;
         return;
      }

      if(local_hour == 12)
      {
         pmflag = 1;
         return;
      }

// It must be afternoon
      local_hour = local_hour - 12;
      pmflag = 1;
   }
}
//===================================================
/*
From http://mathforum.org/dr.math/faq/faq.calendar.html
The Key Value Method

This method uses codes for different months and years to speed up the calculation of the day of the week. You might even be able to memorize the codes.
We'll use December 16, 2482 as an example.

    Take the last 2 digits of the year. In our example, this is 82.

    Divide by 4, and drop any remainder. 82 / 4 = 20, remainder 2, so we think "20."

    Add the day of the month. In our example, 20 + 16 = 36.

    Add the month's key value, from the following table.

    Jan Feb Mar Apr May June July Aug Sept Oct Nov Dec
     1   4   4   0   2   5    0    3   6    1   4   6

    The month for our example is December, with a key value of 6. 36 + 6 = 42.

    If your date is in January or February of a leap year, subtract 1. We're using December, so we don't have to worry about this step.

    Add the century code from the following table. (These codes are for the Gregorian calendar. The rule's slightly simpler for Julian dates.)

    1700s 1800s 1900s 2000s
    4      2      0    6

    Our example year is 2482, and the 2400s aren't in the table. Luckily, the Gregorian calendar repeats every four hundred years. All we have to do is add or subtract 400 until we have a date that is in the table. 2482 - 400 = 2082, so we look at the table for the 2000s, and get the code 6. Now we add this to our running total: 42 + 6 = 48.

    Add the last two digits of the year. 48 + 82 = 130.

    Divide by 7 and take the remainder. This time, 1 means Sunday, 2 means Monday, and so on. A remainder of 0 means Saturday. 130 / 7 = 18, remainder 4, so December 16, 2482 will be on the fourth day of the week-- Wednesday.
*/
//================================================
/*
 Calculate the day of the week

 From http://mathforum.org/dr.math/faq/faq.calendar.html
 The Key Value Method

 The way I have implemented it here it starts with the year 2000
   and is only good until the year 2400

 With Register Optimization only:
   At a 16 MHz clock it takes about 50us if it is not January or February (does not need to determine if leap year);
   If it is January or February and it is a leap year it takes 72 us
   The MSP430G2xxx does not have a hardware math unit so multiplies and divides are emulated

 0  Sat
 1  Sun
 2  Mon
 3  Tue
 4  Wed
 5  Thu
 6  Fri

 Months January to December are 1 - 12
*/
//=======================
// Get the day of the week for UTC time
unsigned char get_day_utc()
{
   unsigned int day;
   unsigned int month;
   unsigned int year;
   unsigned int year2d;
   unsigned int temp;

// monthkey is in main.h
// const unsigned int monthkey[12] = {1,4,4,0,2,5,0,3,6,1,4,6};  // month's key value

/* For testing
   unsigned char utc_month[2] = "12";   // must be 2 digits
   unsigned char utc_day[2] = "03";     // must be 2 digits
   unsigned char utc_year[4] = "2020";
*/

// Convert GPS ASCII month number into binary
// month =  ((utc_month[0]-0x30) * 10) + (utc_month[1]-0x30);  // month in binary
//   month = (utc_month[1] - 0x30);
//   month = month + ((utc_month[0] - 0x30)) * 10;
// x = y * 10 = y * (8 + 2) = y*8 + y*2 = y<<3 + y<<1
   temp =  utc_month[0]-0x30;
   month = (temp<<3) + (temp<<1);
   month = month + (utc_month[1]- 0x30);   // month in binary

// Convert GPS ASCII day number into binary
// day =  ((utc_day[0]-0x30) * 10) + (utc_day[1]-0x30);  // day in binary
//   day = (utc_day[1] - 0x30);
//   day = day + ((utc_day[0] - 0x30)) * 10;
// x = y * 10 = y * (8 + 2) = y*8 + y*2 = y<<3 + y<<1
   temp =  utc_day[0]-0x30;
   day = (temp<<3) + (temp<<1);
   day = day + (utc_day[1]- 0x30);   // day in binary

// Convert GPS ASCII year number into full binary and last two digits
//   year2d = (utc_year[3]- 0x30);
//   year2d = year2d + ((utc_year[2]- 0x30)) * 10;
   temp =  utc_year[2]-0x30;
   year2d = (temp<<3) + (temp<<1);
   year2d = year2d + (utc_year[3]- 0x30);   // last two digits of year in binary

   year = year2d + ((utc_year[1]- 0x30)) * 100;
   year = year + (utc_year[0]-0x30) * 1000;

   temp = year2d>>2;   // last two digits of the year; divide by 4; drop the remainder
   temp = temp + day;  // Add the day of the month
   temp = temp + monthkey[month-1];  // Add the month's key value, from the table.

// If the date is in January or February of a leap year, subtract 1.
   if( (month == 1 ) | (month == 2) )
   {
      if( is_leap2(year) )  { temp = temp-1;}
   }

   temp = temp + 6;       // add the century code (good until 2400); 0 for 1900s
   temp = temp + year2d;  // Add the last two digits of the year
   temp = temp % 7;       // divide by 7; use the remainder which is the day of the week

   asm(" nop");    // for emulator

   return(temp);
}
//===============================================
// Get day of the week for local time
unsigned char get_day_local2()
{
   unsigned int year2d;
   unsigned int temp;

// monthkey is in main.h
// const unsigned int monthkey[12] = {1,4,4,0,2,5,0,3,6,1,4,6};  // month's key value

// Convert GPS ASCII year number into full binary and last two digits
//   year2d = (utc_year[3]- 0x30);
//   year2d = year2d + ((utc_year[2]- 0x30)) * 10;
// x = y * 10 = y * (8 + 2) = y*8 + y*2 = y<<3 + y<<1
   temp =  utc_year[2]-0x30;
   year2d = (temp<<3) + (temp<<1);
   year2d = year2d + (utc_year[3]- 0x30);   // last two digits of year in binary

   temp = year2d>>2;   // last two digits of the year; divide by 4; drop the remainder
   temp = temp + local_day;  // Add the day of the month
   temp = temp + monthkey[local_month-1];  // Add the month's key value, from the table.

// If the date is in January or February of a leap year, subtract 1.
   if( (local_month == 1 ) | (local_month == 2) )
   {
      if( is_leap() )  { temp = temp-1;}
   }

   temp = temp + 6;  // add the century code (good until 2400); 0 for 1900s
   temp = temp + year2d;  // Add the last two digits of the year
   temp = temp % 7;  // divide by 7; use the remainder which is the day of the week

   asm(" nop");    // for emulator

   return(temp);
}
//================================================
// Is it a leap year?
//   The year must be evenly divisible by 4
//   But not if it is evenly divisible by 100 unless it is evenly divisible by 400

unsigned char is_leap()
{
   unsigned int Lyear;

// Convert GPS ASCII to binary
   Lyear = (utc_year[0]-0x30) * 1000;
   Lyear = Lyear + ((utc_year[1]- 0x30)) * 100;
   Lyear = Lyear + ((utc_year[2]- 0x30)) * 10;
   Lyear = Lyear + (utc_year[3]- 0x30);             // years in binary

//   Lyear = 2100;    // for testing

   if( (Lyear & 0x003) != 0)
   { return(0);}                 // it is not evenly divisible by 4
// it is evenly divided by 4

   if(Lyear % 400 == 0)   // it is evenly divisible by 400 (remainder is zero)
   { return(1);}
//  it was not evenly divisible by 400

   if(Lyear % 100 == 0)  {return(0);}   // it is otherwise evenly divisible by 100

// it was evenly divisible by 4, not a century year unless it was 400 year century
   return(1);
}
// Is it a leap year?
//   The year must be evenly divisible by 4
//   But not if it is evenly divisible by 100 unless it is evenly divisible by 400
//================================================
unsigned char is_leap2(unsigned int year)    // years in binary
{
   if( (year & 0x003) != 0)
   { return(0);}                 // it is not evenly divisible by 4
// it is evenly divided by 4

   if(year % 400 == 0)   // it is evenly divisible by 400 (remainder is zero)
   { return(1);}
//  it was not evenly divisible by 400

   if(year % 100 == 0)  {return(0);}   // it is otherwise evenly divisible by 100

// it was evenly divisible by 4, not a century year unless it was 400 year century
   return(1);
}
//==========================================================
//==========================================================
void wait(unsigned int time)
{
    volatile unsigned int i,j;

    for(i=time; i>0; i--)
    {
        for(j=382; j>0; j--);
        {
        }
    }

    return;
}
//==========================================================
// wait 10 ms
void wait10ms(void)
{
   wait(41);
}
//==========================================================
void wait4(unsigned int time)        // 500 uS
{
    volatile unsigned int i,j;

    for(i=time; i>0; i--)
    {
        for(j=20; j>0; j--);
        {
        }
    }

    return;
}
//==========================================================
void waitsec()
{
   int n;

   for(n=200; n>0; n--)
   {wait10ms();}
}
//==========================================================
//==========================================================
void lcd_init(void)
{
    lcd_reset();
}
//===================================
void lcd_reset(void)
{
/*
P2.0 - LCD D4
P2.1 - LCD D5
P2.2 - LCD D6
P2.3 - LCD D7

P2.4 - LCD E (Enable High, Write Data, Enable Low)
P2.5 - LCD RS (Register Select; RS low = Instruction, RS High = Data)
 */

// Wait at least 16 ms after power-up
    wait10ms();    // wait 10 ms
    wait10ms();    // wait 10 ms

    P2OUT &= (~BIT5);    // RS = 0

    lcd_wr4(0x03);      // write nybble and wait 200us
    wait10ms();

    lcd_wr4(0x03);      // write nybble and wait 200us
    lcd_wr4(0x03);      // write nybble and wait 200us
    lcd_wr4(0x02);      // write nybble and wait 200us
    lcd_wr_instr(0x28); //  2 lines, 5x7 dots
    lcd_wr_instr(0x08); //  display off
    lcd_wr_instr(0x01); //  display on

    wait10ms();

//    lcd_wr_instr(0x0F); //  display on, cursor on, blink on
    lcd_wr_instr(0x0c);  //  display on, cursor oFF, blink oFF

    lcd_wr_instr(0x06);  //  entry mode set: cursor increment, no shift
}
//===============================================================
// Do LCD Message
void domsg(const struct messages *sptr)
{
    unsigned char m,n;
    unsigned char data;

    m = sptr->pos;  // position
    lcd_wr_instr(m);

    m = sptr->numchars;  // numbers of chars

    for(n=0; n<m; n++)
    {
        data = sptr->msg[n];
        lcd_wr_data(data);
    }
}
//=================================================================
void lcd_wr4(unsigned int data)   // write nybble and wait 200us
{
//    P2.4 - LCD E (Enable High, Write Data, Enable Low)
//    P2.5 - LCD RS (Register Select; RS low = Instruction, RS High = Data)

    P2OUT |= BIT4;     // Enable High

    data &= 0x0F;
    P2OUT &= 0xF0;
    P2OUT |= data;     // write nybble

    P2OUT &= (~BIT4);  //  /E low

    wait4(20);         // 200 us
}
//======================
// LCD - Write 8-bit instruction
// P2.5 - LCD RS (Register Select; RS low = Instruction, RS High = Data)
void lcd_wr_instr(unsigned int data)
{
   unsigned int dsave;

   P2OUT &= (~BIT5);    // RS = 0

   dsave = data;
   dsave  = dsave >>4;       // high nybble
   lcd_wr4(dsave);  // write nybble and wait 200 us

   data &=0xF;      // low nybble4
   lcd_wr4(data);
}
//======================
// LCD - Write 8-bit data
// P2.5 - LCD RS (Register Select; RS low = Instruction, RS High = Data)
void lcd_wr_data(unsigned int data)
{
    unsigned int dsave;

    P2OUT |= BIT5;    // RS = 1

    dsave = data;
    dsave = dsave >>4;  // high nybble
    lcd_wr4(dsave);     // write nybble and wait 200 us

    data &=0xF;         // low nybble4
    lcd_wr4(data);
}
//==================================================================
// Clear one line
void clear_lcd(unsigned char pos)
{
    unsigned int n;

    lcd_wr_instr(pos);

    for(n=16; n>0; --n)
    { lcd_wr_data(' ');}
}
//==================================================================
void lcd_wr_bcd(unsigned int pos, unsigned int bin)
{
// First Convert Binary to ASCII
// Adapted from 8-bit microcontroller code using a DAA instruction

// unsigned int bin = 0x386; // 0x67 = 103d; 0x386 = 902d
   unsigned char bcd0 = 0;
   unsigned char bcd1 = 0;
   unsigned char bcd2 = 0;
   unsigned char char1, char2, char3;
   int i;
   unsigned char hcarry;

   for(i=0; i<12; i++)
   {
//---------- BCD0
      bcd0 = bcd0<<1;

      if(bin & 0x800)
      {bcd0 = bcd0 +1;}

      bin = bin<<1;

      if(bcd0 >9)
      {
         bcd0 = bcd0 - 10;
         hcarry = 1;
      }
      else {hcarry = 0; }

//---------- BCD1
      bcd1 = bcd1<<1;
      bcd1 = bcd1 + hcarry;

      if(bcd1 >9)
      {
          bcd1 = bcd1 - 10;
          hcarry = 1;
      }
      else {hcarry = 0;}

//---------- BCD2
      bcd2 = bcd2<<1;
      bcd2 = bcd2 + hcarry;

     if(bcd2 >9)
     {
         bcd2 = bcd2 - 10;
//       bcd3 = bcd3 + 1;
//       hcarry = 1;
     }
//        else {hcarry = 0;}

   }

// Convert to ascii
   char1 = bcd2 + 0x30;
   char2 = bcd1 + 0x30;
   char3 = bcd0 + 0x30;

// Display it
   if(char1 == '0')  { char1 = ' '; }    // remove leading '0'

   lcd_wr_instr(pos);
   lcd_wr_data(char1);
   lcd_wr_data(char2);
   lcd_wr_data(char3);
   lcd_wr_data(' ');
   lcd_wr_data('m');
   lcd_wr_data('s');
}
//====================================================================
void lcd_wr_bcd_long(unsigned int pos, unsigned long bin)
{
// First Convert Binary to ASCII
// Adapted from 8-bit microcontroller code with a DAA instruction

//bin = 12345678;    // test

unsigned char bcd0 = 0;
unsigned char bcd1 = 0;
unsigned char bcd2 = 0;
unsigned char bcd3 = 0;
unsigned char bcd4 = 0;
unsigned char bcd5 = 0;
unsigned char bcd6 = 0;
unsigned char bcd7 = 0;

//unsigned char char1, char2, char3, char4, char5, char6, char7, char8;
unsigned char char4, char5, char6, char7, char8;
int i;
unsigned char hcarry;

   for(i=0; i<32; i++)
   {
//---------- BCD0
      bcd0 = bcd0<<1;

      if(bin & 0x80000000)
      {bcd0 = bcd0 +1;}

      bin = bin<<1;

      if(bcd0 >9)
      {
         bcd0 = bcd0 - 10;
         hcarry = 1;
   }
   else {hcarry = 0; }

//---------- BCD1
      bcd1 = bcd1<<1;
      bcd1 = bcd1 + hcarry;

      if(bcd1 >9)
      {
         bcd1 = bcd1 - 10;
         hcarry = 1;
      }
      else {hcarry = 0;}

//---------- BCD2
      bcd2 = bcd2<<1;
      bcd2 = bcd2 + hcarry;

     if(bcd2 >9)
     {
         bcd2 = bcd2 - 10;
         hcarry = 1;
     }
        else {hcarry = 0;}

//---------- BCD3
        bcd3 = bcd3<<1;
        bcd3 = bcd3 + hcarry;

       if(bcd3 >9)
       {
           bcd3 = bcd3 - 10;
           hcarry = 1;
       }
          else {hcarry = 0;}

//---------- BCD4
       bcd4 = bcd4<<1;
       bcd4 = bcd4 + hcarry;

       if(bcd4 >9)
       {
          bcd4 = bcd4 - 10;
          hcarry = 1;
       }
       else {hcarry = 0;}

//---------- BCD5
       bcd5 = bcd5<<1;
       bcd5 = bcd5 + hcarry;

       if(bcd5 >9)
       {
          bcd5 = bcd5 - 10;
          hcarry = 1;
       }
       else {hcarry = 0;}

//---------- BCD6
       bcd6 = bcd6<<1;
       bcd6 = bcd6 + hcarry;

       if(bcd6 >9)
       {
          bcd6 = bcd6 - 10;
          hcarry = 1;
       }
       else {hcarry = 0;}

//---------- BCD7
       bcd7 = bcd7<<1;
       bcd7 = bcd7 + hcarry;

       if(bcd7 >9)
       {
          bcd7 = bcd7 - 10;
          hcarry = 1;
        }
       else {hcarry = 0;}
   }

// Convert to ascii
//   char1 = bcd7 + 0x30;
//   char2 = bcd6 + 0x30;
//   char3 = bcd5 + 0x30;
   char4 = bcd4 + 0x30;
   char5 = bcd3 + 0x30;
   char6 = bcd2 + 0x30;
   char7 = bcd1 + 0x30;
   char8 = bcd0 + 0x30;

// Display it
   lcd_wr_instr(pos);
//   lcd_wr_data(char1);
//   lcd_wr_data(char2);
//   lcd_wr_data(' ');
//   lcd_wr_data('.');

//   lcd_wr_data(char3);

   if(char4 == '0')
   {
      char4 = ' ';

      if(char5 == '0')
      {
         char5 = ' ';

         if(char6 == '0')
         {
            char6 = ' ';
         }

      }

   }

   lcd_wr_data(char4);
   lcd_wr_data(char5);
//   lcd_wr_data(' ');
//   lcd_wr_data('.');

   lcd_wr_data(char6);
   lcd_wr_data(char7);
   lcd_wr_data(char8);

   lcd_wr_data(' ');
   lcd_wr_data('H');
   lcd_wr_data('z');
}
//=================================================================
unsigned char convert_hex_bcd2(signed char bin, unsigned char *outstring)
{
// Adapted from 8-bit microcontroller code with a DAA instruction

   unsigned char bcd0 = 0;
   unsigned char bcd1 = 0;
   int i;
   unsigned char hcarry, flag;

   if(bin<0)
   {
       bin = -bin;
       flag = 1;   // was negative number
   }
   else { flag = 0;}  // positive number

   for(i=0; i<8; i++)
   {
//---------- BCD0
      bcd0 = bcd0<<1;

      if(bin & 0x80)
      {bcd0 = bcd0 +1;}

      bin = bin<<1;

      if(bcd0 >9)
      {
         bcd0 = bcd0 - 10;
         hcarry = 1;
      }
      else {hcarry = 0; }

//---------- BCD1
      bcd1 = bcd1<<1;
      bcd1 = bcd1 + hcarry;

      if(bcd1 >9)
      {
         bcd1 = bcd1 - 10;
         hcarry = 1;
      }
      else {hcarry = 0;}
   }

// Convert to ascii
//   outstring[0] = bcd1 + 0x30;
//   outstring[1] = bcd0 + 0x30;

   outstring[0] = bcd1 + 0x30;
   outstring[1] = bcd0 + 0x30;

   return(flag);
}
//==================================================
void convert_hex_bcd4(unsigned int bin, unsigned char *outstring)
{
// Adapted from 8-bit microcontroller code with a DAA instruction
   unsigned char bcd0 = 0;
   unsigned char bcd1 = 0;
   unsigned char bcd2 = 0;
   unsigned char bcd3 = 0;
   int i;
   unsigned char hcarry;

   for(i=0; i<16; i++)
   {
//---------- BCD0
      bcd0 = bcd0<<1;

      if(bin & 0x8000)
      {bcd0 = bcd0 +1;}

      bin = bin<<1;

      if(bcd0 >9)
      {
         bcd0 = bcd0 - 10;
         hcarry = 1;
      }
      else {hcarry = 0; }

//---------- BCD1
      bcd1 = bcd1<<1;
      bcd1 = bcd1 + hcarry;

      if(bcd1 >9)
      {
         bcd1 = bcd1 - 10;
         hcarry = 1;
      }
      else {hcarry = 0;}

//---------- BCD2
      bcd2 = bcd2<<1;
      bcd2 = bcd2 + hcarry;

      if(bcd2 >9)
      {
         bcd2 = bcd2 - 10;
         hcarry = 1;
      }
      else {hcarry = 0;}

//---------- BCD3
      bcd3 = bcd3<<1;
      bcd3 = bcd3 + hcarry;

      if(bcd3 >9)
      {
          bcd3 = bcd3 - 10;
          hcarry = 1;
      }
      else {hcarry = 0;}

   }  // end of for loop

// Convert to ascii
   outstring[0] = bcd3 + 0x30;
   outstring[1] = bcd2 + 0x30;
   outstring[2] = bcd1 + 0x30;
   outstring[3] = bcd0 + 0x30;
}
//==========================================================
//==========================================================
void do_switch(unsigned char port, unsigned char bit, struct myswitch *sptr)    // PortIn, Bit, switch structure
{
// Debounce and process the switch
// sptr points to the switch's data block
    sptr->debounce  = sptr->debounce <<1;    // shifts in a zero
    if(port & bit)        // a zero has already been shifted in
    {
       sptr->debounce++ ;      // so make LSB a one
    }

// Debounced state is either 0x00 or 0xFF
    if(sptr->debounce == 0x00)  {sptr->state = 0x00;}
    if(sptr->debounce == 0xFF)  {sptr->state = 0x01;}

// process the switch
    if(sptr->state != sptr->old)     // we have a change in switch state
    {
       if( (sptr->old == 0x01) && (sptr->state == 0x00) )  // we have a switch press
       {
          sptr->press = 0x01;
       }

       sptr->old = sptr->state;     // the new state is now the old state
    }

    asm(" nop");    // for emulator
}
//=================================================
//=================================================
void display_local()   // current (local) time
{
   unsigned int weekday, m;
   unsigned char temp;

   time12();   // Adjust if 12 hour time

// Convert binary to BCD and then to ASCI
   convert_hex_bcd2(local_hour,loc_hour_ptr);
   convert_hex_bcd2(local_minute,loc_min_ptr);
   convert_hex_bcd2(local_second,loc_sec_ptr);
   convert_hex_bcd2(local_month,loc_month_ptr);
   convert_hex_bcd2(local_day,loc_day_ptr);
   convert_hex_bcd4(local_year,loc_year_ptr);

   lcd_wr_instr(0x80);        // position of time
   lcd_wr_data(loc_hour[0]);
   lcd_wr_data(loc_hour[1]);

   lcd_wr_data(':');
   lcd_wr_data(loc_minute[0]);
   lcd_wr_data(loc_minute[1]);

   lcd_wr_data(':');
   lcd_wr_data(loc_second[0]);
   lcd_wr_data(loc_second[1]);

   lcd_wr_instr(0xC0);        // position of date
   lcd_wr_data(loc_month[0]);
   lcd_wr_data(loc_month[1]);

   lcd_wr_data('/');
   lcd_wr_data(loc_day[0]);
   lcd_wr_data(loc_day[1]);

   lcd_wr_data('/');
   lcd_wr_data(loc_year[0]);
   lcd_wr_data(loc_year[1]);
   lcd_wr_data(loc_year[2]);
   lcd_wr_data(loc_year[3]);

   weekday = get_day_local2();     // day of the week

   lcd_wr_instr(0xCC);
   for(m=0;m<3;m++)
   {
      temp =  day_array[weekday][m];
      lcd_wr_data(temp);
   }

   if(!hours12) {domsg(&msg3);}    // "Local"
   else
   {
      if(pmflag)  {domsg(&msg13);}  // "PM"
      else {domsg(&msg12);}         // "AM"
   }

   asm(" nop");    // for emulator
}
//===============================================
// Get day of the week for local time
unsigned char get_day_local()
{
   unsigned int year;
   unsigned int year2d;
   unsigned int temp;
   signed char Lday, Lmonth;

// monthkey is in main.h
// const unsigned int monthkey[12] = {1,4,4,0,2,5,0,3,6,1,4,6};  // month's key value

// Convert ASCII year number into full binary and last two digits
   year2d = (loc_year[3]- 0x30);
   year2d = year2d + ((loc_year[2]- 0x30)) * 10;
   year = year2d + ((loc_year[1]- 0x30)) * 100;
   year = year + (loc_year[0]-0x30) * 1000;

   Lday = (loc_day[1] - 0x30);
   Lday = Lday + ((loc_day[0] - 0x30)) * 10;

   Lmonth = (loc_month[1] - 0x30);
   Lmonth = Lmonth + ((loc_month[0] - 0x30)) * 10;

   temp = year2d>>2;   // last two digits of the year; divide by 4; drop the remainder
   temp = temp + Lday;  // Add the day of the month
   temp = temp + monthkey[Lmonth-1];  // Add the month's key value, from the table.


// If the date is in January or February of a leap year, subtract 1.
   if( (Lmonth == 1 ) | (Lmonth == 2) )
   {
      if( is_leap2(year) )  { temp = temp-1;}
   }

   temp = temp + 6;  // add the century code (good until 2400); 0 for 1900s
   temp = temp + year2d;  // Add the last two digits of the year
   temp = temp % 7;  // divide by 7; use the remainder which is the day of the week

   asm(" nop");    // for emulator

   return(temp);
}
//==============================
// Get day of the week for the day (between 2000 and 2400)
unsigned char get_day(int month, int day, int year)
{
   unsigned int year2d;
   unsigned int temp;
   unsigned char year_bcd[4];

// monthkey is in main.h
// const unsigned int monthkey[12] = {1,4,4,0,2,5,0,3,6,1,4,6};  // month's key value

// convert binary year to BCD
   convert_hex_bcd4(year, year_bcd);   // get the new local hour in ASCII

// Convert ASCII year number to last two digits
   year2d = (year_bcd[3]- 0x30);
   year2d = year2d + ( (year_bcd[2]- 0x30) ) * 10;

   temp = year2d>>2;   // last two digits of the year; divide by 4; drop the remainder
   temp = temp + day;  // Add the day of the month
   temp = temp + monthkey[month-1];  // Add the month's key value, from the table.

// If the date is in January or February of a leap year, subtract 1.
   if( (month == 1) | (month == 2) )
   {
      if( is_leap2(year) )  { temp = temp-1;}
   }

   temp = temp + 6;  // add the century code (good until 2400); 0 for 1900s
   temp = temp + year2d;  // Add the last two digits of the year
   temp = temp % 7;  // divide by 7; use the remainder which is the day of the week

   asm(" nop");    // for emulator

   return(temp);
}
//================================================
void find_dst(unsigned int year)
{
// DST starts on the second Sunday in March
// Find day of the week for March 1
   unsigned char day;

   day = get_day(3,1,year);   // 0 is Saturday
   dst_start_date = dst_start_table[day];      // use a table

   day = get_day(11,1,year);                  // use a table
   dst_end_date = dst_end_table[day];
}
//==================================================
// Are we in DST?
unsigned char is_dst()
{
   if(!dst_observed)    // If DST is not observed then we already know the answer is No
   { return(0); }

// Are we past March and before November?
   if( (local_month >3) & (local_month<11))   // we are definitely in the DST window
   { return(1); }

// Are we before March or after November?
   if( (local_month <3) | (local_month>11))   // we are definitely not in the DST window
   { return(0); }

// If we are in March are we before the start date and time of DST for the year?
   if( (local_month == 3) )
   {
      if( (local_day < dst_start_date) ) { return(0); }   // not in the window
      if( (local_day > dst_start_date) ) { return(1); }   // in the window

      // we are in the day when 2AM becomes 3AM?
      if( (local_hour < 02) )
      { return(0); }                // not in the window
      if( (local_hour >= 02) )
      { return(1); }                // We are in the window
      //  Since the time changed from 2AM to 3AM we don't have to remember it changed
   }

// If we are in November are we before the end date and time of DST for the year?
   if( (local_month == 11) )
   {
      if( (local_day < dst_end_date) ) { return(1); }   // still in the window
      if( (local_day > dst_end_date) ) { return(0); }   // not in the window

      // this is the day and we have to know if we have already changed it or 1AM-2AM will be wrong
      if( (local_hour < 01) )  { return(1); }    // still in the window
      if( (local_hour > 02) )  { return(0); }    // not in the window

      // we are now between 1AM and 2AM
//      if(!dst_window) { return(0);}  // we already turned the window off

      // now we are waiting for 2 AM or later
      if( (local_hour >= 02) ) { return(0); }   // We are no longer in the window (and will remember it in dst_window which is global)
   }

   asm(" nop");    // for emulator

   return(0);   // should never get here
}
//==================================================
//==================================================
void write_SegC(char F_utc_offset, char F_dst_observed, char F_hours12)
{
   unsigned char counth, countl;
   unsigned int count;    // number of write cycles
   char *Flash_ptr;       // Flash pointer

   Flash_ptr = (char *) 0x1040; // Initialize Flash pointer

   counth = *(Flash_ptr+6);     // count of program cycles, do it all in bytes
   countl = *(Flash_ptr+7);
   count = (counth<<8) + countl;

   if(count == 0xFFFF)   { count = 1;}
   else  {count = count + 1;}

   counth = count>>8;
   countl = count & 0xFF;

   FCTL2 = FWKEY + FSSEL0 + 44;  // MCLK 16MHz/44 = 363KHz for Flash Timing Generator

// Erase INFO C
   FCTL1 = FWKEY + ERASE;          // Set Erase bit
   FCTL3 = FWKEY;                  // Clear Lock bit
   *Flash_ptr = 0;                 // Dummy write to erase Flash segment

// Write 6 bytes: utc_offset, utc_offset, dst_observe, dst-observe, counth, countl
   FCTL1 = FWKEY + WRT;            // Set WRT bit for write operation

   *Flash_ptr++ = F_utc_offset;    // Write value to flash
   *Flash_ptr++ = F_utc_offset;    // Write value to flash (twice for redundancy)

   *Flash_ptr++ = F_dst_observed;  // Write value to flash
   *Flash_ptr++ = F_dst_observed;  // Write value to flash (twice for redundancy)

   *Flash_ptr++ = F_hours12;       // Write value to flash
   *Flash_ptr++ = F_hours12;       // Write value to flash (twice for redundancy)

   *Flash_ptr++ = counth;          // Write value to flash - keep a count of the write cycles
   *Flash_ptr++ = countl;          // Write value to flash

   FCTL1 = FWKEY;                  // Clear WRT bit
   FCTL3 = FWKEY + LOCK;           // Set LOCK bit

   wr_flash_flag = 0;              // we have done it
}
//===================================
void write_SegC_test (char F_utc_offset, char F_dst_observed, char F_hours12)
{
   unsigned char counth, countl;
   unsigned int count;    // number of write cycles
   char *Flash_ptr;       // Flash pointer

   Flash_ptr = (char *) 0x1040; // Initialize Flash pointer

   counth = *(Flash_ptr+6);     // do it all in bytes
   countl = *(Flash_ptr+7);
   count = (counth<<8) + countl;

   if(count == 0xFFFF)   { count = 1;}
   else  {count = count + 1;}

   counth = count>>8;
   countl = count & 0xFF;

   wr_flash_flag = 0;              // we have done it
}
//===================================
void erase_SegC()  // erase
{
   char *Flash_ptr;       // Flash pointer

   Flash_ptr = (char *) 0x1040; // Initialize Flash pointer

   FCTL2 = FWKEY + FSSEL0 + 43;  // MCLK 16MHz/44 = 363KHz for Flash Timing Generator

// Erase INFO C
   FCTL1 = FWKEY + ERASE;          // Set Erase bit
   FCTL3 = FWKEY;                  // Clear Lock bit
   *Flash_ptr = 0;                 // Dummy write to erase Flash segment

   FCTL1 = FWKEY;                  // Clear WRT bit
   FCTL3 = FWKEY + LOCK;           // Set LOCK bit
}
//===================================================
// Convert GPS ASCII second into binary and add 1 second for the lcd delay
void convert_utc_bin()
{
    unsigned int temp;

// ut_second = ( (utc_sec[0]-0x30) * 10) + (utc_sec[1]-0x30);
//   ut_sec = (utc_sec[1] - 0x30);
//   ut_sec = ut_sec + ((utc_sec[0] - 0x30)) * 10;
// x = y * 10 = y * (8 + 2) = y*8 + y*2 = y<<3 + y<<1
   temp =  utc_sec[0]-0x30;
   ut_second = (temp<<3) + (temp<<1);
   ut_second = ut_second + (utc_sec[1]- 0x30);   // second in binary

// ut_minute = ( (utc_min[0]-0x30) * 10) + (utc_min[1]-0x30);
   temp =  utc_min[0]-0x30;
   ut_minute = (temp<<3) + (temp<<1);
   ut_minute = ut_minute + (utc_min[1]- 0x30);   // minute in binary

// ut_hour  = ( (utc_hour[0]-0x30) * 10) + (utc_hour[1]-0x30);  // hours in binary
   temp =  utc_hour[0]-0x30;
   ut_hour = (temp<<3) + (temp<<1);
   ut_hour = ut_hour + (utc_hour[1]- 0x30);   // hour in binary

// ut_month = ( (utc_month[0]-0x30) * 10) + (utc_month[1]-0x30); // month in binary
   temp =  utc_month[0]-0x30;
   ut_month = (temp<<3) + (temp<<1);
   ut_month = ut_month + (utc_month[1]- 0x30);   // month in binary

//   ut_day   = ( (utc_day[0]-0x30) * 10) + (utc_day[1]-0x30);     // day in binary
   temp =  utc_day[0]-0x30;
   ut_day = (temp<<3) + (temp<<1);
   ut_day = ut_day + (utc_day[1]- 0x30);   // day in binary

   ut_year = (utc_year[0]-0x30) * 1000;
   ut_year = ut_year + ((utc_year[1]- 0x30)) * 100;
   ut_year = ut_year + ((utc_year[2]- 0x30)) * 10;
   ut_year = ut_year + (utc_year[3]- 0x30);   // years in binary

   leapyear = is_leap();   // is a leap year?

// We have to add one second to account for the 1 second delay in LCD
   ut_second = ut_second + 1;   // If we wanted to we could add 2 seconds instead
   if(ut_second < 60)    // this is all we have to do
   {
      return;
   }

// add a minute
   ut_second = ut_second - 60;
   ut_minute = ut_minute + 1;
   if(ut_minute < 60)    // this is all we have to do
   {
      return;
   }

// add an hour
   ut_minute = ut_minute - 60;
   ut_hour = ut_hour + 1;
   if(ut_hour < 24)    // this is all we have to do
   {
      return;
   }

// now we have to add a day
   ut_hour = 0;
   ut_day = ut_day + 1;   // add a day

// The days in the months depends on whether it is a leap year
   if(leapyear)
   {
      LastDay = months_ly[ut_month];  // the last day of the month for leap years
   }
   else
   {
      LastDay = months[ut_month];    // the last day of the month for regular years
   }

   if(ut_day <= LastDay)  // we don't have to adjust months and maybe years
   {
      return;
   }

// shazbot, the day was past the month so we have to adjust the month
   ut_hour = 0;
   ut_day = 1;
   ut_month = ut_month + 1;

   if(ut_month < 13)    // year didn't change
   {
      return;
   }

   ut_hour = 0;
   ut_month = 1;     // New year

// and increment the year
   ut_year = ut_year + 1;

// Is the new year a leap year?
   leapyear = is_leap();   // is a leap year?
   asm(" nop");    // for emulator

// end of adding 1 second
   return;
}
//==============================
// Convert utc binary back to ascii and display it
void convert_utc_ascii()
{
//   unsigned int weekday, m;
//   unsigned char temp;

// Convert binary data back into ASCII
   convert_hex_bcd2(ut_second, utc_sec_ptr);
   convert_hex_bcd2(ut_minute, utc_min_ptr);
   convert_hex_bcd2(ut_hour, utc_hour_ptr);
   convert_hex_bcd2(ut_day, utc_day_ptr);
   convert_hex_bcd2(ut_month, utc_mon_ptr);
   convert_hex_bcd4(ut_year, utc_year_ptr);

   asm(" nop");    // for emulator
}
//==================================================
void do_sounds()
{
   if(sound)       // finish the sound trigger
   {
      sound = 0;
      P1OUT |= BIT4;    // Flag 1
      asm(" nop");      // for emulator
   }

   newsec1 = P1IN & BIT0;                   // GPS PPS
   if( (newsec1 == 1) && (oldsec1 == 0) )   // got a transition
   {
      asm(" nop");      // for emulator

      if( (last_utc_sec[0] == '5') & (last_utc_sec[1] == '6') )  // compensate for 1 second for LCD
      {
         P1OUT &= ~BIT5;     // Flag 2: Tell Sound it is second 57
      }
      else
      {
         P1OUT |= BIT5;      // not second 57
      }

      P1OUT &= ~BIT4;   // trigger the sound

      sound = 1;        // finish the sound trigger in the next 4ms interrupt
      asm(" nop");      // for emulator
   }

   oldsec1 = newsec1;
}
//====================================
